import random
import time
from numpy import array
import numpy as np

import pandas as pd
import os
import tensorflow as tf
from keras.models import Sequential
from keras.layers import LSTM
from keras.layers import Dense
from keras.layers import Dropout
from keras.layers import Flatten
from keras.layers import RepeatVector
from keras.layers import TimeDistributed
from keras.layers.convolutional import Conv1D
from keras.layers.convolutional import MaxPooling1D
from sklearn.preprocessing import MinMaxScaler
import matplotlib.pyplot as plt


seed_value = 1234

# Set seed in all environments to help reproducibility
def set_all_seeds(seed_value):
    # 1. Set `PYTHONHASHSEED` environment variable at a fixed value
    os.environ['PYTHONHASHSEED']=str(seed_value)
    # 2. Set `python` built-in pseudo-random generator at a fixed value
    random.seed(seed_value)
    # 3. Set `numpy` pseudo-random generator at a fixed value
    np.random.seed(seed_value)
    # 4. Set `tensorflow` pseudo-random generator at a fixed value
    tf.random.set_seed(seed_value)


# import data

# Set current working directory one layer up
path_parent = os.path.dirname(os.getcwd())
os.chdir(path_parent)

ts_global = np.array(pd.read_csv(os.getcwd() + r"\Data\new_cases_global.csv"))
ts_norway = np.array(pd.read_csv(os.getcwd() + r"\Data\new_cases_Norway.csv"))


def split_sequence(sequence, n_steps):
    # Formats the data for supervised learning
    X, y = list(), list()
    for i in range(len(sequence)):
        # find the end of this pattern
        end_ix = i + n_steps
        # check if we are beyond the sequence
        if end_ix > len(sequence ) -1:
            break
        # gather input and output parts of the pattern
        seq_x, seq_y = sequence[i:end_ix], sequence[end_ix]
        X.append(seq_x)
        y.append(seq_y)
    return array(X), array(y)

def k_step_prediction(k, x_train, model, scaler, prev_days = 7):
    '''
    Iteratively performes a one step prediction and use the newly
    predicted as input to forecast the next day
    :param k: amount of days ahead to predict
    :param x_train: the time series used
    :param model: pre-trained model
    :param scaler: the pre-trained min-max scaler,
    used to revert the prediction back to normal scale
    :param prev_days: Amount of previous days used to predict the next day.
    :return: k-step prediction
    '''
    predictions = []
    x_input = x_train[-prev_days:] # last prev_days days
    x_input = x_input.reshape((1, prev_days, 1))
    yhat = model.predict(x_input, verbose=0)[0][0].reshape(1,1)
    predictions.append(yhat)
    for i in range(1, k):
        x_input = x_input.reshape(prev_days, 1)
        # remove first day and add prediction to the end of the input
        x_input = np.concatenate((x_input[1:], yhat), axis = 0)
        x_input = x_input.reshape((1, prev_days, 1))
        yhat = model.predict(x_input, verbose=0)[0][0].reshape(1,1)
        predictions.append(yhat)

    for i in range(len(predictions)):
        predictions[i] = scaler.inverse_transform(predictions[i])[0][0]
    return predictions


def train_model(X, y, days_ahead, prev_days = 7):
    '''
    :param X: Data matrix
    :param y: Response vector
    :param days_ahead: days ahead to be forecasted after last input
    :param n_steps: amount of previous days used as input in forecasting
    :return:
    '''
    model = Sequential()
    model.add(Conv1D(filters=64, kernel_size=3, activation='relu', input_shape=(prev_days, 1)))
    model.add(Conv1D(filters=64, kernel_size=3, activation='relu'))
    model.add(MaxPooling1D(pool_size=1, strides=1))
    model.add(Dropout(0.2, seed=seed_value))
    model.add(Flatten())
    model.add(RepeatVector(1)) # days_ahead
    model.add(LSTM(200, return_sequences=True))
    model.add(TimeDistributed(Dense(50))) # 100
    model.add(TimeDistributed(Dense(1)))
    model.compile(optimizer='adam', loss='mse') # 'mean_absolute_percentage_error'
    # fit model
    history = model.fit(X, y, epochs=100, verbose=0, batch_size=22)  # skal være 472, 22
    loss = history.history['loss']
    # print(model.summary()) # uncomment to view amount of variables
    return model, loss


def prediction_global(ts, endpoint, prev_days, len_train = -1, days_ahead = 7):
    # The new version uses the repeat vector layer, and uses the
    # Time Distributed Wrapper at the end insted of at the beginning.

    '''
    :param ts: Input time series to be forecasted
    :param endpoint: Last index used for training the model
    :param prev_days: Amount of days used to predict the next days
    :param len_train: Amount of days back from endpoint used to train the model
    :param days_ahead: Amount of days forecasted after endpoint
    :return: days_ahead-forecast
    '''
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    # only use train data  to fit scaler:
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    X, y = split_sequence(x_train, prev_days)
    # reshape from [samples, timesteps] into
    # [samples, subsequences, timesteps, features]
    X = X.reshape((X.shape[0], prev_days, 1))

    preds = np.zeros((10, days_ahead))
    loss = []
    for i in range(10):
        start_time = time.time()
        model, loss_i = train_model(X, y, days_ahead, prev_days=prev_days)
        loss.append(loss_i)
        print("Time used to fit model", i+1, ": ", (time.time() - start_time), " sec")
        start_time = time.time()
        preds[i] = np.round_(k_step_prediction(
            days_ahead, x_train, model, scaler, prev_days=prev_days))
        print("Time used to forecast with model", i + 1, ": ",
              (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/glob_preds_" + str(days_ahead) +
               "_days_ahead_from_" + str(endpoint) + "_with_" + str(len_train) +
               "_train_data.csv", preds)

    # This part can be uncommented to inspect and save the mean loss function
    # loss = array(loss)
    # mean_loss = loss.mean(axis = 0)
    # print(mean_loss.shape)
    # np.savetxt(os.getcwd() +
    #            r"\Predictions/mean_loss_on_original_test_data.csv", mean_loss)
    # print(mean_loss.shape)
    # plt.plot(mean_loss)
    # plt.ylabel('loss')
    # plt.xlabel('epoch')
    # plt.show()
    return preds

def prediction_norway(ts, endpoint, prev_days, len_train = -1, days_ahead = 7):
    # The new version uses the repeat vector layer, and uses the
    # Time Distributed Wrapper at the end insted of at the beginning.
    # Thus, the model traines on a seven day output?

    '''
    :param ts: Input time series to be forecasted
    :param endpoint: Last index used for training the model
    :param prev_days: Amount of days used to predict the next days
    :param len_train: Amount of days back from endpoint used to train the model
    :param days_ahead: Amount of days forecasted after endpoint
    :return: days_ahead-forecast
    '''
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)  # 17th July
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    X, y = split_sequence(x_train, prev_days)
    # reshape from [samples, timesteps] into [samples, subsequences, timesteps, features]
    X = X.reshape((X.shape[0], prev_days, 1))

    preds = np.zeros((10, days_ahead))
    for i in range(10):
        start_time = time.time()
        model, loss = train_model(X, y, days_ahead, prev_days=prev_days)
        print("Time used to fit model", i+1, ": ", (time.time() - start_time), " sec")
        start_time = time.time()
        preds[i] = np.round_(k_step_prediction(
            days_ahead, x_train, model, scaler, prev_days=prev_days))
        print("Time used to forecast with model", i + 1, ": ",
              (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/nor_preds_" + str(days_ahead) +
               "_days_ahead_from_" + str(endpoint) + "_with_" + str(len_train) +
               "_train_data.csv", preds)
    return preds


def single_pred_with_test_data(days_ahead, x_train_and_test, model, scaler, prev_days = 7):
    '''
    Performes a series of one step predictions,
    one step at a time to generate a k-step prediction.
    At each iteration, the newly observed day is
    incorporated into the input of the prediction.
    :return: k-step prediction of covid-19 new cases
    '''
    predictions = []
    # how far back to start the input?
    length_relevant_data = prev_days + days_ahead -1

    # the first input to the prediction is the last
    # prev_days elements of the train data
    for i in range(0, days_ahead-1):
        x_input = x_train_and_test[-length_relevant_data + i:
                                   -length_relevant_data + prev_days + i]
        x_input = x_input.reshape((1, prev_days, 1))
        yhat = model.predict(x_input, verbose=0)[0][0].reshape(1, 1)
        predictions.append(yhat)

    x_input = x_train_and_test[-prev_days:]
    x_input = x_input.reshape((1, prev_days, 1))
    yhat = model.predict(x_input, verbose=0)[0][0].reshape(1, 1)
    predictions.append(yhat)

    for i in range(len(predictions)):
        predictions[i] = scaler.inverse_transform(predictions[i])[0][0]
    return predictions

def one_step_preds_global(ts, endpoint, prev_days, len_train = -1, days_ahead = 7):
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    # last len_train days used to predict
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    # test data on day d is used to predict new cases on day d+1
    x_train_and_test = ts[endpoint - len_train:
                          (endpoint+days_ahead-1)].reshape(len_train+days_ahead-1, 1)
    x_train_and_test = scaler.transform(x_train_and_test)
    # split train data into input-output matrix
    X, y = split_sequence(x_train, prev_days)
    X = X.reshape((X.shape[0], prev_days, 1))

    # Make 10 predictions and take mean forecast, to reduce variability in forecast
    preds = np.zeros((10, days_ahead))

    for i in range(10):
        start_time = time.time()
        # The model is only trained in the train data!
        model, loss = train_model(X, y, days_ahead, prev_days=prev_days)
        print("Time used to fit model", i+1, ": ", (time.time() - start_time), " sec")
        start_time = time.time()
        preds[i] = np.round_(single_pred_with_test_data(days_ahead, x_train_and_test,
                                                        model, scaler, prev_days=prev_days))
        print("Time used to forecast with model", i+1, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/glob_test_preds_" + str(days_ahead) +
               "_days_ahead_from_" + str(endpoint) + "_with_" + str(len_train) + "_train_data.csv", preds)
    return preds

def one_step_preds_norway(ts, endpoint, prev_days, len_train = -1, days_ahead = 7):
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1) # last len_train days used to predict
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    # test data on day d is used to predict new cases on day d+1
    x_train_and_test = ts[endpoint - len_train:(endpoint+days_ahead-1)].reshape(len_train+days_ahead-1, 1)
    x_train_and_test = scaler.transform(x_train_and_test)
    # split train data into input-output matrix
    X, y = split_sequence(x_train, prev_days)
    X = X.reshape((X.shape[0], prev_days, 1))

    # Make 10 predictions and take mean forecast, to reduce variability in forecast
    preds = np.zeros((10, days_ahead))
    for i in range(10):
        start_time = time.time()
        model, loss = train_model(X, y, days_ahead, prev_days=prev_days) # The model is only trained in the train data!
        print("Time used to fit model", i+1, ": ", (time.time() - start_time), " sec")
        start_time = time.time()
        preds[i] = np.round_(single_pred_with_test_data(days_ahead, x_train_and_test, model, scaler, prev_days=prev_days))
        print("Time used to forecast with model", i+1, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/nor_test_preds_" + str(days_ahead) +
               "_days_ahead_from_" + str(endpoint) + "_with_" + str(len_train) + "_train_data.csv", preds)
    return preds

def one_step_preds_on_train_global(ts, endpoint, prev_days, len_train = -1, days_ahead = 7):
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1) # last len_train days used to predict
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    # test data on day d is used to predict new cases on day d+1
    x_train_and_test = ts[endpoint - len_train:(endpoint+days_ahead-1)].reshape(len_train+days_ahead-1, 1)
    x_train_and_test = scaler.transform(x_train_and_test)
    # split train data into input-output matrix
    X, y = split_sequence(x_train, prev_days)
    X = X.reshape((X.shape[0], prev_days, 1))
    # Make 10 predictions and take mean forecast, to reduce variability in forecast
    preds = np.zeros((10, days_ahead))
    for i in range(10):
        start_time = time.time()
        model, loss = train_model(X, y, days_ahead, prev_days=prev_days) # The model is only trained in the train data!
        print("Time used to fit model", i+1, ": ", (time.time() - start_time), " sec")
        start_time = time.time()
        preds[i] = np.round_(single_pred_with_test_data(days_ahead, x_train_and_test[:-days_ahead], model, scaler, prev_days=prev_days))
        print("Time used to forecast with model", i+1, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/glob_test_preds_train_" + str(days_ahead) +
               "_from_" + str(endpoint) + "_with_" + str(len_train) + "_train_data.csv", preds)
    return preds

def one_step_preds_on_train_norway(ts, endpoint, prev_days, len_train = -1, days_ahead = 7):
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1) # last len_train days used to predict
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    # test data on day d is used to predict new cases on day d+1
    x_train_and_test = ts[endpoint - len_train:(endpoint+days_ahead-1)].reshape(len_train+days_ahead-1, 1)
    x_train_and_test = scaler.transform(x_train_and_test)
    # split train data into input-output matrix
    X, y = split_sequence(x_train, prev_days)
    X = X.reshape((X.shape[0], prev_days, 1))
    # Make 10 predictions and take mean forecast, to reduce variability in forecast
    preds = np.zeros((10, days_ahead))
    for i in range(10):
        start_time = time.time()
        model, loss = train_model(X, y, days_ahead, prev_days=prev_days) # The model is only trained in the train data!
        print("Time used to fit model", i+1, ": ", (time.time() - start_time), " sec")
        start_time = time.time()
        preds[i] = np.round_(single_pred_with_test_data(days_ahead, x_train_and_test[:-days_ahead], model, scaler, prev_days=prev_days))
        print("Time used to forecast with model", i+1, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/nor_test_preds_on_train_" + str(days_ahead) +
               "_from_" + str(endpoint) + "_with_" + str(len_train) + "_train_data.csv", preds)
    return preds

def all_in_one_global(endpoint, days_ahead, len_train):
    # Wrapper function to systemize predictions results
    # The function will output both the one-step prediction and the days_ahead prediction
    prediction_global(ts=ts_global, endpoint=endpoint, prev_days=7, days_ahead=days_ahead, len_train=len_train)
    one_step_preds_global(ts=ts_global, endpoint=endpoint, prev_days=7, days_ahead = days_ahead, len_train = len_train)


def all_in_one_norway(endpoint, days_ahead, len_train):
    # Wrapper function to systemize predictions,
    # The function will output both the one-step prediction and the days_ahead prediction
    one_step_preds_norway(ts=ts_norway, endpoint=endpoint, prev_days=7, days_ahead = days_ahead, len_train = len_train)
    prediction_norway(ts=ts_norway, endpoint=endpoint, prev_days=7, days_ahead = days_ahead, len_train = len_train)

def accuracy_with_simulations_global(ts, simulations, endpoint, days_ahead, prev_days = 7, len_train = -1):
    '''
    Forecasts with scheme 1, but use simulations instead of observations
    :param ts: The time series of interest
    :param simulations: Simulated observations from Gandalf model, imported from Matlab
    :param endpoint: Last index of dataset used in training
    :param prev_days: how many days put in to the model (7 is default)
    :param days_ahead: days ahead forecast
    :return: the RRMSEs and MAPEs associated with each simulation
    '''
    set_all_seeds(seed_value)

    simulations = np.array(simulations)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)  # last len_train days used to predict
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    # split train data into input-output matrix
    X, y = split_sequence(x_train, prev_days)
    X = X.reshape((X.shape[0], prev_days, 1))

    # Make 10 predictions and take mean forecast, to reduce variability in forecast
    RRMSEs = np.zeros(simulations.shape[0])
    MAPEs = np.zeros(simulations.shape[0])

    models = [] # This saves a lot of time!
    for j in range(10):
        model, loss = train_model(X, y, days_ahead,
                                  prev_days=prev_days)  # The model is only trained in the train data!
        models.append(model)

    for i in range(simulations.shape[0]):
        start_time = time.time()
        simulation = scaler.transform(simulations[i].reshape(-1, 1))
        input_in_prediction = np.concatenate((x_train, simulation[0:-1]), axis=0)
        ten_preds = np.zeros((10, days_ahead))
        for j in range(10):
            ten_preds[j] = np.round_(
                single_pred_with_test_data(days_ahead, input_in_prediction, models[j], scaler, prev_days=prev_days))
        mean_pred = np.mean(ten_preds, axis = 0)

        simulation = scaler.inverse_transform(simulation.reshape(-1, 1))
        mean_sim = simulation.mean()
        MSE = np.square(np.subtract(simulation, mean_pred)).mean()
        RRMSEs[i] = np.round_(np.sqrt(MSE) / mean_sim * 100, decimals=3)
        mape = tf.keras.losses.MeanAbsolutePercentageError()
        MAPEs[i] = np.round_(mape(simulation, mean_pred).numpy(), decimals=3)
        print("Time used to predict with simulation", i, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/RRMSEs_from_simulation_global_" + str(days_ahead) +
               "_days_ahead_with_" + str(len_train) + "_train_data_from" + str(endpoint) + ".csv", RRMSEs)
    np.savetxt(os.getcwd() + r"\Predictions/MAPEs_from_simulation_global_" + str(days_ahead) +
               "_days_ahead_with_" + str(len_train) + "_train_data_from" + str(endpoint) + ".csv", MAPEs)


def predictions_with_simulations_global(ts, simulations, endpoint, days_ahead, prev_days = 7, len_train = -1):
    '''
    Forecasts with scheme 1, but use simulations instead of observations, the accuracy measures are returned
    :param ts: The time series of interest
    :param simulations: Simulated observations from Gandalf model, imported from Matlab
    :param endpoint: Last index of dataset used in training
    :param prev_days: how many days put in to the model (7 is default)
    :param days_ahead: days ahead forecast
    :return: the RRMSEs and MAPEs associated with each simulation
    '''
    set_all_seeds(seed_value)

    simulations = np.array(simulations)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)  # last len_train days used to predict
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    # split train data into input-output matrix
    X, y = split_sequence(x_train, prev_days)
    X = X.reshape((X.shape[0], prev_days, 1))

    # Make 10 predictions and take mean forecast, to reduce variability in forecast
    preds_matlab = np.zeros((3, 28))

    start_time = time.time()

    models = []  # This saves a lot of time!
    for j in range(10):
        model, loss = train_model(X, y, days_ahead,
                                  prev_days=prev_days)  # The model is only trained in the train data!
        models.append(model)
    print("time to fit ten models: ", time.time() - start_time)

    for i in range(3):  # for the first three predictions
        preds = np.zeros((10, 28))
        print("simulations[i]", simulations[i])
        simulation = scaler.transform(simulations[i].reshape(-1, 1))
        input_in_prediction = np.concatenate((x_train, simulation[0:-1]), axis=0)
        for j in range(10):
            pred = np.round_(
                single_pred_with_test_data(days_ahead, input_in_prediction, models[j], scaler, prev_days=prev_days))
            preds[j] = pred
        preds_matlab[i] = np.mean(preds, axis = 0)
    print("Time used to predict all simulations: ", (time.time() - start_time), " sec")

    print(preds_matlab, preds_matlab.shape)
    np.savetxt(os.getcwd() + r"\Predictions/pred_three_preds_on_sims.csv", preds_matlab)


def predict_weekday_no_test(Xs, Ys, x_trains, days_ahead, scaler, prev_days = 7):
    '''
    The function predicts a days_ahead forecast for seven models, one for each weekday.
    :param datas: list of seven scaled train data, one containing all Mondays etc...
    :param days_ahead: how many days ahead to predict
    :param scaler: pre-trained scaler function
    :param prev_days: how many days used to predict preceding day for each submodel
    :return: days_ahead forecast
    '''

    prediction = np.zeros(days_ahead)
    prediction_days = [i for i in range(days_ahead)] # sequence of numbers
    for i in range(len(Xs)):
        days_to_predict = prediction_days[i::7] # divides what days to predict by each submodel, the length is the amount of days to predict
        X = Xs[i]
        y = Ys[i]
        x_train = x_trains[i]
        model, loss = train_model(X, y, len(days_to_predict), prev_days)
        prediction[days_to_predict] = k_step_prediction(len(days_to_predict), x_train, model, scaler, prev_days) # prediction of all of a single weekday
    return prediction




def weekday_models_no_test_data_global(ts, endpoint, days_ahead, prev_days = 7, len_train = -1):
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)  # use train data to fit scaler
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    x_train_1 = x_train[::7] # Every 7th day
    x_train_2 = x_train[1::7]
    x_train_3 = x_train[2::7]
    x_train_4 = x_train[3::7]
    x_train_5 = x_train[4::7]
    x_train_6 = x_train[5::7]
    x_train_7 = x_train[6::7]
    print(x_train_1.shape, x_train_2.shape, x_train_3.shape, x_train_4.shape, x_train_5.shape, x_train_6.shape,x_train_7.shape)

    x_trains = [x_train_1, x_train_2, x_train_3, x_train_4, x_train_5, x_train_6, x_train_7]
    X_1, y_1 = split_sequence(x_train_1, prev_days)
    X_2, y_2 = split_sequence(x_train_2, prev_days)
    X_3, y_3 = split_sequence(x_train_3, prev_days)
    X_4, y_4 = split_sequence(x_train_4, prev_days)
    X_5, y_5 = split_sequence(x_train_5, prev_days)
    X_6, y_6 = split_sequence(x_train_6, prev_days)
    X_7, y_7 = split_sequence(x_train_7, prev_days)

    Ys = [y_1, y_2, y_3, y_4, y_5, y_6, y_7]
    # reshape from [samples, timesteps] into [samples, subsequences, timesteps, features]
    X_1 = X_1.reshape((X_1.shape[0], prev_days, 1))
    X_2 = X_2.reshape((X_2.shape[0], prev_days, 1))
    X_3 = X_3.reshape((X_3.shape[0], prev_days, 1))
    X_4 = X_4.reshape((X_4.shape[0], prev_days, 1))
    X_5 = X_5.reshape((X_5.shape[0], prev_days, 1))
    X_6 = X_6.reshape((X_6.shape[0], prev_days, 1))
    X_7 = X_7.reshape((X_7.shape[0], prev_days, 1))
    Xs =[X_1, X_2, X_3, X_4, X_5, X_6, X_7]
    preds = np.zeros((10, days_ahead))

    for i in range(10):
        start_time = time.time()
        preds[i] = np.round_(predict_seven_no_test(Xs, Ys, x_trains, days_ahead, scaler, prev_days))
        print("Time used to fit and forecast with model", i + 1, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/glob_weekday_preds_" + str(days_ahead) +
               "_days_ahead_from_" + str(endpoint) + "_with_" + str(len_train) + "_train_data.csv", preds)

def predict_weekday_with_test(Xs, Ys, x_train_and_tests, days_ahead, scaler, prev_days = 7):
    '''
    The function predicts a days_ahead forecast for seven models, one for each weekday.
    :param datas: list of seven scaled train data, one containing all Mondays etc...
    :param days_ahead: how many days ahead to predict
    :param scaler: pre-trained scaler function
    :param prev_days: how many days used to predict preceding day for each submodel
    :return: days_ahead forecast
    '''

    prediction = np.zeros(days_ahead)
    prediction_days = [i for i in range(days_ahead)] # sequence of numbers
    for i in range(len(Xs)):
        days_to_predict = prediction_days[i::7] # divides what days to predict by each submodel, the length is the amount of days to predict
        X = Xs[i]
        y = Ys[i]
        x_train_and_test = x_train_and_tests[i]
        model, loss = train_model(X, y, len(days_to_predict), prev_days)
        # prediction of all of a single weekday:
        prediction[days_to_predict] = \
            single_pred_with_test_data(len(days_to_predict), x_train_and_test, model, scaler, prev_days)
    return prediction

def weekday_models_with_test_data_global(ts, endpoint, days_ahead, prev_days = 7, len_train = -1):
    set_all_seeds(seed_value)
    if len_train == -1:
        len_train = endpoint
    scaler = MinMaxScaler()
    x_train = ts[endpoint - len_train:endpoint].reshape(len_train, 1)  # use train data to fit scaler
    scaler.fit(x_train)
    x_train = scaler.transform(x_train)
    x_train_1 = x_train[::7] # Every 7th day
    x_train_2 = x_train[1::7]
    x_train_3 = x_train[2::7]
    x_train_4 = x_train[3::7]
    x_train_5 = x_train[4::7]
    x_train_6 = x_train[5::7]
    x_train_7 = x_train[6::7]

    x_train_and_test = ts[endpoint - len_train:(endpoint+days_ahead-1)].reshape(len_train+days_ahead-1, 1)
    x_train_and_test = scaler.transform(x_train_and_test)
    x_train_and_test_1 = x_train_and_test[::7] # Every 7th day
    x_train_and_test_2 = x_train_and_test[1::7]
    x_train_and_test_3 = x_train_and_test[2::7]
    x_train_and_test_4 = x_train_and_test[3::7]
    x_train_and_test_5 = x_train_and_test[4::7]
    x_train_and_test_6 = x_train_and_test[5::7]
    x_train_and_test_7 = x_train_and_test[6::7]

    x_train_and_tests = [x_train_and_test_1, x_train_and_test_2, x_train_and_test_3, x_train_and_test_4,
                x_train_and_test_5, x_train_and_test_6, x_train_and_test_7]

    X_1, y_1 = split_sequence(x_train_1, prev_days)
    X_2, y_2 = split_sequence(x_train_2, prev_days)
    X_3, y_3 = split_sequence(x_train_3, prev_days)
    X_4, y_4 = split_sequence(x_train_4, prev_days)
    X_5, y_5 = split_sequence(x_train_5, prev_days)
    X_6, y_6 = split_sequence(x_train_6, prev_days)
    X_7, y_7 = split_sequence(x_train_7, prev_days)

    Ys = [y_1, y_2, y_3, y_4, y_5, y_6, y_7]
    # reshape from [samples, timesteps] into [samples, subsequences, timesteps, features]
    X_1 = X_1.reshape((X_1.shape[0], prev_days, 1))
    X_2 = X_2.reshape((X_2.shape[0], prev_days, 1))
    X_3 = X_3.reshape((X_3.shape[0], prev_days, 1))
    X_4 = X_4.reshape((X_4.shape[0], prev_days, 1))
    X_5 = X_5.reshape((X_5.shape[0], prev_days, 1))
    X_6 = X_6.reshape((X_6.shape[0], prev_days, 1))
    X_7 = X_7.reshape((X_7.shape[0], prev_days, 1))
    Xs =[X_1, X_2, X_3, X_4, X_5, X_6, X_7]
    preds = np.zeros((10, days_ahead))

    for i in range(10):
        start_time = time.time()
        preds[i] = np.round_(predict_weekday_with_test(Xs, Ys, x_train_and_tests, days_ahead, scaler, prev_days))
        print("Time used to fit and forecast with model", i + 1, ": ", (time.time() - start_time), " sec")

    np.savetxt(os.getcwd() + r"\Predictions/glob_weekday_preds_test_" + str(days_ahead) +
               "_days_ahead_from_" + str(endpoint) + "_with_" + str(len_train) + "_train_data.csv", preds)




############################### Global ###################################

# 28 day predictions global


###############
all_in_one_global( endpoint = 40, days_ahead = 28, len_train = 40)
# all_in_one_global( endpoint = 196, days_ahead = 28, len_train = 196)
# all_in_one_global( endpoint = 196, days_ahead = 28, len_train = 50)
# all_in_one_global( endpoint = 196, days_ahead = 28, len_train = 100)
#
# ##############
# all_in_one_global(endpoint = 100, days_ahead = 28, len_train = 100)
# all_in_one_global(endpoint = 300, days_ahead = 28, len_train = 300)
# all_in_one_global(endpoint = 500, days_ahead = 28, len_train = 500)
# all_in_one_global(endpoint = 743, days_ahead = 28, len_train = 743) # newest
#

# # Fewer observations: How robust is the model?
# all_in_one_global(endpoint = 100, days_ahead = 28, len_train = 50)
# all_in_one_global(endpoint = 300, days_ahead = 28, len_train = 50)
# all_in_one_global(endpoint = 300, days_ahead = 28, len_train = 100)
# all_in_one_global(endpoint = 500, days_ahead = 28, len_train = 50)
# all_in_one_global(endpoint = 500, days_ahead = 28, len_train = 100)
# all_in_one_global(endpoint = 743, days_ahead = 28, len_train = 50)
# all_in_one_global(endpoint = 743, days_ahead = 28, len_train = 100)

############################### Norway ###################################

# An equivalent set of predictions can be made on the norwegian data set.
# Start at 264 to compare with original Gandalf model.

# all_in_one_norway(endpoint = 264, days_ahead = 28, len_train = 264)
# all_in_one_norway(endpoint = 264, days_ahead = 28, len_train = 50)
# all_in_one_norway(endpoint = 264, days_ahead = 28, len_train = 100)
#
# all_in_one_norway(endpoint = 100, days_ahead = 28, len_train = 100)
# all_in_one_norway(endpoint = 300, days_ahead = 28, len_train = 300)
# all_in_one_norway(endpoint = 500, days_ahead = 28, len_train = 500)
# all_in_one_norway(endpoint = 696, days_ahead = 28, len_train = 696)
#
# # Fewer observations: How robust is the model?
# all_in_one_norway(endpoint = 100, days_ahead = 28, len_train = 50)
# all_in_one_norway(endpoint = 300, days_ahead = 28, len_train = 50)
# all_in_one_norway(endpoint = 300, days_ahead = 28, len_train = 100)
# all_in_one_norway(endpoint = 500, days_ahead = 28, len_train = 50)
# all_in_one_norway(endpoint = 500, days_ahead = 28, len_train = 100)
# all_in_one_norway(endpoint = 696, days_ahead = 28, len_train = 50)
# all_in_one_norway(endpoint = 696, days_ahead = 28, len_train = 100)


########################## Simulation study #############################

simulations = pd.read_csv(
    os.getcwd() + r"\Matlab\Saved variables\simulations_from_gandalf.csv", header=None)


# print("simulation.shape = ", simulations.shape)
#print("simulations.iloc[0:2, :] = ", simulations.iloc[0:3, :])
# accuracy_with_simulations_global(ts_global, simulations, endpoint=196, days_ahead=28)
# predictions_with_simulations_global(ts_global, simulations.iloc[0:3, :], endpoint=196, days_ahead=28)
# predictions_with_simulations_global(ts_norway, simulations.iloc[0:3, :], endpoint=196, days_ahead=28)

######################### Predict with weekday model  ######################

# weekday_models_no_test_data_global(ts = ts_global, endpoint = 196, days_ahead = 7)
# weekday_models_with_test_data_global(ts = ts_global, endpoint = 600, days_ahead = 7)